import csv
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

from .file_utils import secure_filename
from .utils import get_temp_file_path, get_bytes_and_delete, TempFile, get_zip_response_from_bytes, get_bytes


def get_csv_bytes(data):
    file_path = get_temp_file_path()
    f = open(file_path, 'w')
    wr = csv.writer(f, quoting=csv.QUOTE_ALL)
    for row in data:
        wr.writerow(row)
    f.close()
    return get_bytes_and_delete(file_path)


def get_csv_temp_file(data):
    f = NamedTemporaryFile('w')
    wr = csv.writer(f, quoting=csv.QUOTE_ALL)
    for row in data:
        wr.writerow(row)
    f.flush()
    return TempFile(f)


def get_csv_package_response_from_sheets(sheets, file_name):
    with NamedTemporaryFile() as tmp_zip_file:
        with ZipFile(tmp_zip_file.name, "a") as out_zip_file:
            for (sheet_name, sheet) in sheets:
                with get_csv_temp_file(sheet) as csv_file:
                    out_zip_file.write(csv_file.name, secure_filename(f'{sheet_name}.csv'))
        return get_zip_response_from_bytes(get_bytes(tmp_zip_file.name), file_name)



THEME_COLORS = {}


def clamp(val, minimum=0, maximum=255):
    if val < minimum:
        return minimum
    if val > maximum:
        return maximum
    return val


def colorscale(color_hex, scale_factor):
    """
    Scales a hex string by ``scale_factor``. Returns scaled hex string.

    To darken the color, use a float value between 0 and 1.
    To brighten the color, use a float value greater than 1.

    >>> colorscale("#DF3C3C", .5)
    #6F1E1E
    >>> colorscale("#52D24F", 1.6)
    #83FF7E
    >>> colorscale("#4F75D2", 1)
    #4F75D2
    """

    color_hex = color_hex.strip('#')

    if scale_factor < 0 or len(color_hex) != 6:
        return color_hex

    r, g, b = int(color_hex[:2], 16), int(color_hex[2:4], 16), int(color_hex[4:], 16)

    r = clamp(r * scale_factor)
    g = clamp(g * scale_factor)
    b = clamp(b * scale_factor)

    return "#%02x%02x%02x" % (r, g, b)


def generate_admin_theme_colors():
    from django.conf import settings

    theme_overrides = getattr(settings, 'REUSABLE_ADMIN_THEME_OVERRIDE', {})
    if theme_overrides:
        try:
            global THEME_COLORS
            if 'background_color' in theme_overrides:
                THEME_COLORS.update(dict((f'bgColor{x + 1}',
                                          colorscale(theme_overrides['background_color'], opacity))
                                         for x, opacity in enumerate([1, 2, 3])))
            if 'link_color' in theme_overrides:
                link_color = theme_overrides['link_color']
                THEME_COLORS.update(dict(
                    linkColor=link_color,
                    linkHoverColor=colorscale(link_color, 2),
                    linkVisitedColor=colorscale(link_color, 3)))
            THEME_COLORS.update(dict(
                textColor=theme_overrides.get('text_color', '')
            ))
        except Exception as e:
            print(e)



from django.urls import reverse, NoReverseMatch
from django.utils.safestring import mark_safe

DEFAULT_SITE = 'admin'


class ModelURLs:
    def __init__(self, model, site=DEFAULT_SITE, pk=None):
        self.model = model
        self.app_label, self.model_name = self.model._meta.app_label, self.model._meta.model_name
        self.pk = pk
        self.site = site

    def get_add_url_name(self):
        return '%s:%s_%s_add' % (self.site, self.app_label, self.model_name)

    def get_add_url(self):
        return reverse(self.get_add_url_name())

    def get_changelist_url_name(self):
        return '%s:%s_%s_changelist' % (self.site, self.app_label, self.model_name)

    def get_changelist_url(self):
        return reverse(self.get_changelist_url_name())

    def get_change_url_name(self):
        return '%s:%s_%s_change' % (self.site, self.app_label, self.model_name)

    def _get_change_url(self, pk):
        return reverse(self.get_change_url_name(), args=[pk])

    def get_obj_change_link_tag(self, text, css_class='', target=''):
        return f'<a target="{target}" class="{css_class}" href="{self.get_obj_change_url()}">{text}</a>'

    def get_obj_change_url(self):
        return self._get_change_url(self.pk)

    def get_delete_url_name(self):
        return '%s:%s_%s_delete' % (self.site, self.app_label, self.model_name)

    def _get_delete_url(self, pk):
        return reverse(self.get_delete_url_name(), args=[pk])

    def get_obj_delete_url(self):
        return self._get_delete_url(self.pk)

    def get_change_link(self):
        if self.pk:
            try:
                return mark_safe(u'<a href="{w}" class="admin-edit-link">Details</a>'.format(
                    w=self.get_obj_change_url()))
            except NoReverseMatch:
                pass
        return None



import os

from django import forms
from django.utils.safestring import mark_safe

from django_reusable.utils import xstr


class FileLinkWidget(forms.TextInput):
    def __init__(self, base_path, base_url, attrs=None):
        self.base_path = base_path
        self.base_url = base_url
        super(FileLinkWidget, self).__init__(attrs)

    def get_link(self, value):
        if value and os.path.exists(os.path.join(self.base_path, value)):
            return mark_safe('<a href="%s" target="_blank" class="btn btn-default">Open</a>' %
                             os.path.join(self.base_url, value))
        return ''

    def render(self, name, value, attrs=None, renderer=None):
        html = '''
            <div class="input-group">
                %s
                <span class="input-group-btn"> 
                    %s
                </span>
            </div>
        ''' % (super().render(name, value, attrs, renderer), self.get_link(value))
        return html


class ReadonlySelect(forms.Select):
    def render(self, name, value, attrs=None, renderer=None):
        matches = [t for (v, t) in self.choices if v == value]
        text = matches[0] if matches else None
        return f'''{text or '---'}<input name="{name}" value="{value}" type="hidden" />'''


class ReadonlyMultiSelect(forms.SelectMultiple):
    def render(self, name, value, attrs=None, renderer=None):
        value_list = value if isinstance(value, list) or isinstance(value, tuple) else eval(value)
        matches = [t for (v, t) in self.choices if v in value_list]
        text = ', '.join(matches) if matches else None
        attrs = attrs or {}
        attrs['style'] = 'display: none;'
        default_widget = super().render(name, value, attrs, renderer)
        return f'''{text or '---'}{default_widget}'''


class ReadOnlyInput(forms.TextInput):
    def render(self, name, value, attrs=None, renderer=None):
        attrs = attrs or {}
        attrs['style'] = 'display: none;'
        return f'{xstr(value)}{super().render(name, value, attrs, renderer)}'


class DateInput(forms.DateInput):
    input_type = 'date'


class SingleCharSplitInput(forms.TextInput):
    def __init__(self, split=2, *args, **kwargs):
        self.split = split
        super().__init__(*args, **kwargs)

    def render(self, name, value, attrs=None, renderer=None):
        attrs = attrs or {}
        split_input_attrs = attrs.copy()
        split_input_attrs.update({
            'minlength': 1,
            'maxlength': 1,
            'style': 'width: 20px; margin-right: 3px;',
            'class': 'form-control split-char',
            'id': '',
        })
        attrs_str = " ".join([f'{k}="{v}"' for (k, v) in split_input_attrs.items()])
        inputs = [f'<input {attrs_str} />' for x in range(0, self.split)]
        attrs['style'] = 'display: none;'
        return '''
                <script>
                    $(document).ready(function () {
                        $('input.split-char').on("keyup", function(){
                            if($(this).val().length == $(this).attr("maxlength")){
                                $(this).next().focus();
                                $("#id_%s").val(Array.from($('input.split-char')).map(x => x.value).join(''));
                            }
                        }); 
                    })
                </script>
               ''' % (name, ) + f'''    
            <div class='single-char-split-inputs'>
                <div>{"".join(inputs)}</div>
                {super().render(name, value, attrs, renderer)}
            </div>
        '''


import itertools

from django import forms
from django.utils.safestring import mark_safe
from django_tables2 import CheckBoxColumn as DefaultCheckBoxColumn, Column
from django_tables2.utils import AttributeDict

from django_reusable.utils import global_request, humane_currency, format_as_currency
from django_reusable.forms.widgets import ReadOnlyInput


class EnhancedColumn(Column):
    def __init__(self,
                 new_row_index: int = None,
                 colspan: int = None,
                 no_empty_cell: bool = False,
                 *args, **kwargs):
        self.new_row_index = new_row_index
        self.no_empty_cell = no_empty_cell
        super().__init__(*args, **kwargs)
        if colspan:
            td_attrs = self.attrs.get('td', {})
            td_attrs['colspan'] = colspan
            self.attrs['td'] = td_attrs
        if self.has_new_row_index:
            self.visible = False

    @property
    def has_new_row_index(self):
        return self.new_row_index is not None and self.new_row_index > 0


class TextFieldColumn(EnhancedColumn):
    empty_values = []

    def render(self, value):
        ff = forms.CharField(widget=forms.Textarea)
        ff.widget.attrs['class'] = 'form-control'
        return ff.widget.render(self.verbose_name, value)


class EnhancedCheckBoxColumn(DefaultCheckBoxColumn):
    def select_checked(self, value, record):
        request = global_request()
        selected_ids = dict(request.POST).get(self.bound_column, [])
        return str(value) in selected_ids

    def __init__(self, header=None, attrs=None, checked=None, bound_column=None, **extra):
        self._header = header
        self.bound_column = bound_column
        checked = checked or self.select_checked
        super().__init__(attrs, checked, **extra)

    @property
    def header(self):
        attrs = self.attrs.get("th__input", {})
        attrs["onchange"] = '''$(this).parents('table').find(
        `tr td:nth-child(${$(this).parents("th").index() + 1}) input`
        ).prop("checked", this.checked);
        '''
        self.attrs["th__input"] = attrs
        header = self._header or super().header
        return header


class RadioButtonColumn(DefaultCheckBoxColumn):
    @property
    def header(self):
        return ''

    def render(self, value, bound_column, record):
        default = {"type": "radio", "name": bound_column.name, "value": value}
        if self.is_checked(value, record):
            default.update({"checked": "checked"})

        general = self.attrs.get("input")
        specific = self.attrs.get("td__input")
        attrs = AttributeDict(default, **(specific or general or {}))
        return mark_safe("<input %s/>" % attrs.as_html())


class ChoiceColumn(EnhancedColumn):
    empty_values = []

    def __init__(self, choices, widget=forms.Select, css_class='form-control', *args, **kwargs):
        self.choices = choices
        self.widget = widget
        self.css_class = css_class
        super().__init__(*args, **kwargs)

    def render(self, value):
        ff = forms.ChoiceField(choices=self.choices, widget=self.widget)
        ff.widget.attrs['class'] = self.css_class
        return ff.widget.render(self.verbose_name, value)


class CounterColumn(EnhancedColumn):
    empty_values = []

    def __init__(self, start=0, *args, **kwargs):
        self.start = start
        super().__init__(*args, **kwargs)

    def render(self, value):
        self.row_counter = getattr(self, 'row_counter', itertools.count())
        return self.start + next(self.row_counter)


def add_class(attrs, element, css_class):
    el = attrs.get(element, {})
    el_classes = el.get('class', '').split(' ')
    if css_class not in el_classes:
        el_classes.append(css_class)
    el['class'] = ' '.join(el_classes)
    attrs[element] = el


class NumberColumn(EnhancedColumn):
    def __init__(self, *args, **kwargs):
        attrs = kwargs.get('attrs', {})
        add_class(attrs, 'td', 'text-right')
        add_class(attrs, 'tf', 'text-right')
        kwargs['attrs'] = attrs
        super().__init__(*args, **kwargs)


class CurrencyColumn(NumberColumn):

    def __init__(self, human_format=False, *args, **kwargs):
        self.human_format = human_format
        super().__init__(*args, **kwargs)

    def render(self, value):
        return humane_currency(value) if self.human_format else format_as_currency(value)


class HiddenIdInputColumn(EnhancedColumn):
    def render(self, value):
        field = forms.CharField(widget=ReadOnlyInput())
        return mark_safe(field.widget.render('id', value))



# Generated by Django 2.0.5 on 2022-06-29 15:33

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Error',
            fields=[
                ('hash', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('host', models.CharField(max_length=1024)),
                ('path', models.CharField(max_length=4096)),
                ('method', models.CharField(max_length=64)),
                ('request_data', models.TextField()),
                ('exception_name', models.CharField(max_length=256)),
                ('traceback', models.TextField()),
                ('count', models.IntegerField(default=0)),
                ('created_on', models.DateTimeField(auto_now=True)),
                ('last_seen', models.DateTimeField(auto_now=True, db_index=True)),
            ],
        ),
    ]



from copy import deepcopy

from django.contrib import admin, messages
from django.contrib.admin.options import BaseModelAdmin, InlineModelAdmin
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse, path
from django.utils.safestring import mark_safe

from django_reusable.admin.urls import ModelURLs
from django_reusable.admin.utils import remove_from_fieldsets
from django_reusable.admin.filters import SearchInFilter
from django_reusable.forms.forms import EnhancedBaseInlineFormSet
from django_reusable.utils import ifilter


class EnhancedAdminInlineMixin(InlineModelAdmin):
    select2_inlines_fields = []
    default_field_queryset = dict()
    limit_field_queryset_model_fields = dict()
    limit_saved_queryset_value_fields = []
    formset = EnhancedBaseInlineFormSet

    def get_formset(self, request, obj=None, **kwargs):
        formset = super().get_formset(request, obj, **kwargs)
        formset.limit_saved_queryset_value_fields = self.limit_saved_queryset_value_fields
        formset.default_field_queryset = self.default_field_queryset
        formset.limit_field_queryset_model_fields = self.limit_field_queryset_model_fields
        return formset

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        field = super().formfield_for_dbfield(db_field, request, **kwargs)
        if db_field.name in self.select2_inlines_fields:
            field.widget.attrs['class'] = 'inline-select2'
        return field


class EnhancedBaseAdminMixin(BaseModelAdmin):
    user_field = 'created_by'
    only_delete_owned = False

    def has_delete_permission(self, request, obj=None):
        if self.only_delete_owned and not request.user.has_perm('user.is_admin'):
            if not (obj and getattr(obj, self.user_field).id == request.user.id):
                return False
        return super(EnhancedBaseAdminMixin, self).has_delete_permission(request, obj)


class AjaxActionMixin:
    admin_site = None
    """
        List of tuples:
        ('field_name', dict(btn_text, btn_class, additional_html, callback, short_desc))
    """
    ajax_action_fields = []

    def get_readonly_fields(self, request, obj=None):
        return list(super().get_readonly_fields(request, obj)) + [x[0] for x in self.get_ajax_action_fields()]

    def get_ajax_action_fields(self):
        return self.ajax_action_fields

    def get_callback_key(self, name):
        return f'{self.__class__.__name__}_{name}'

    def register_callback(self, name, callback):
        if not hasattr(self.admin_site, 'ajax_handler_callback_params'):
            self.admin_site.ajax_handler_callback_params = {}
        param_key = self.get_callback_key(name)
        if not self.admin_site.ajax_handler_callback_params.get(param_key):
            self.admin_site.ajax_handler_callback_params[param_key] = (self, callback)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for name, config in self.get_ajax_action_fields():
            btn_text = config.get('btn_text', 'Button')
            btn_class = config.get('btn_class', 'btn btn-warning')

            @mark_safe
            def func(instance):
                if not instance.id:
                    return ''
                url = reverse('django_reusable:ajax_callback_handler',
                              args=(instance.pk, self.get_callback_key(name)))
                return (f'<button class="{btn_class} ajax-action-btn" data-url="{url}">{btn_text}</button>' +
                        config.get('additional_html', ''))

            func.short_description = config.get('short_desc', btn_text)
            func.__name__ = name
            setattr(self, func.__name__, func)

            self.register_callback(name, config['callback'])


class EnhancedAdminMixin(admin.ModelAdmin, EnhancedBaseAdminMixin):
    default_filters = []
    search_in_choices = []
    """
        List of tuples:
        ('field_name', lambda instance: instance.field.name, optional short_desc)
    """
    custom_fields = []
    """
        List of tuples:
        ('name', dict(btn_text, btn_class, stay_on_page, callback))
    """
    extra_change_form_buttons = []
    """
        List of tuples:
        (url, dict(link_text, link_class, new_tab, user_passes_test, pk_passes_test))
    """
    extra_changelist_links = []
    """
        List of tuples:
        ('name', dict(btn_text, btn_class, callback, short_desc, custom_redirect))
        Note: callback is called with args (instance)
    """
    action_links = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.search_in_choices:
            if not ifilter(lambda x: 'SearchInFilter' in str(x), self.list_filter):
                filter_class = type('%sSearchInFilter' % self.__class__.__name__, (SearchInFilter,),
                                    {'lookup_choices': sorted(self.search_in_choices, key=lambda x: x[1])})
                self.list_filter.insert(0, filter_class)
            self.base_search_fields = map(lambda xy: xy[0], self.search_in_choices)

        for f in self.custom_fields:
            func = f[1]
            func.allow_tags = True
            if len(f) > 2:
                func.short_description = f[2]
            func.__name__ = f[0]
            setattr(self, func.__name__, func)

        def get_action_link_fn(name, attrs):
            func = lambda instance: mark_safe(
                '<a class="btn {0}" href="{1}">{2}</a>'.format(
                    attrs.get('btn_class', ''),
                    reverse(f'admin:{self.model._meta.app_label}_{self.model._meta.model_name}_action_view_{name}',
                            args=(instance.id,)),
                    attrs.get('btn_text', ''),
                ))
            func.allow_tags = True
            if 'short_desc' in attrs:
                func.short_description = attrs['short_desc']
            func.__name__ = name
            return func

        for name, attrs in self.action_links:
            if 'callback' not in attrs:
                continue
            func = get_action_link_fn(name, attrs)
            setattr(self, func.__name__, func)

    def get_search_fields(self, request):
        search_fields = deepcopy(self.search_fields)
        if self.search_in_choices:
            search_fields = [request.GET.get(SearchInFilter.parameter_name)] if request.GET.get(
                SearchInFilter.parameter_name) else (search_fields or self.base_search_fields)
        return search_fields

    def hide_save_buttons(self, request, object_id):
        """
        Hook method to hide save buttons and disable POST on some conditions.

        :param request: HttpRequest object
        :param object_id: Object ID of the current change page
        :return: True to hide save buttons, False otherwise
        """
        return False

    def get_changelist_extra_context(self, request):
        return {}

    def custom_changelist_actions(self, request):
        pass

    def _add_extra_change_form_buttons(self, request, pk, extra_context):
        extra_context.update(
            extra_submit_buttons=[
                (f'__{name}', config.get('btn_class', 'btn-primary'), config.get('btn_text', 'Button'))
                for (name, config) in self.extra_change_form_buttons
                if (config.get('user_passes_test', lambda u: True)(request.user) and
                    config.get('pk_passes_test', lambda _pk: True)(pk))
            ]
        )

    def _handle_extra_change_form_buttons(self, request, object_id):
        for (name, config) in self.extra_change_form_buttons:
            if f'__{name}' not in request.POST:
                continue
            callback = config.get('callback')
            if callback:
                r = callback(self, request, object_id)
                if r:
                    messages.info(request, r)
            if config.get('stay_on_page', False):
                return HttpResponseRedirect(
                    ModelURLs(self.model, self.admin_site.name, object_id).get_obj_change_url()
                )
            break
        return None

    def _add_extra_changelist_links(self, request, extra_context):
        extra_context.update(
            extra_links=[
                (url,
                 config.get('link_class', 'btn-link'),
                 config.get('link_text', 'Link'),
                 config.get('new_tab', False))
                for (url, config) in self.get_extra_changelist_links(request)
                if config.get('user_passes_test', lambda u: True)(request.user)
            ]
        )

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        self.custom_changelist_actions(request)
        self._add_extra_changelist_links(request, extra_context)
        if self.default_filters:
            try:
                test = request.META['HTTP_REFERER'].split(request.META['PATH_INFO'])
                if test and test[-1] and not test[-1].startswith('?'):
                    url = reverse('admin:%s_%s_changelist' % (self.opts.app_label, self.opts.model_name))
                    filters = []
                    for f in self.default_filters:
                        key = f.split('=')[0]
                        if key not in request.GET:
                            filters.append(f)
                    if filters:
                        return HttpResponseRedirect("%s?%s" % (url, "&".join(filters)))
            except:
                pass
        extra_context.update(self.get_changelist_extra_context(request) or {})
        return super(EnhancedAdminMixin, self).changelist_view(request, extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        extra_context.update(
            hide_save_buttons=self.hide_save_buttons(request, object_id)
        )
        self._add_extra_change_form_buttons(request, object_id, extra_context)
        if request.method == 'POST' and extra_context['hide_save_buttons']:
            raise PermissionDenied()
        response = super().change_view(request, object_id, form_url, extra_context)
        if request.method == 'POST':
            return self._handle_extra_change_form_buttons(request, object_id) or response
        return response

    def remove_inline_instances(self, inline_instances, *inlines_to_remove):
        to_remove = set()
        for inline_class in inlines_to_remove:
            inline = inline_class(self.model, self.admin_site)
            for index, i in enumerate(inline_instances):
                if type(inline) == type(i):
                    to_remove.add(i)
        for inline in to_remove:
            if inline in inline_instances:
                inline_instances.remove(inline)

    def get_readonly_fields(self, request, obj=None):
        return (list(super().get_readonly_fields(request, obj)) +
                [x[0] for x in self.custom_fields] +
                [x[0] for x in self.action_links])

    def get_extra_changelist_links(self, request):
        return self.extra_changelist_links

    def get_fieldset_section_exclusions(self, request, obj):
        return []

    def get_fieldset_field_exclusions(self, request, obj):
        return []

    def get_fieldsets(self, request, obj=None):
        original_fieldsets = super().get_fieldsets(request, obj)
        field_exclusions = self.get_fieldset_field_exclusions(request, obj)
        section_exclusions = self.get_fieldset_section_exclusions(request, obj)
        if not field_exclusions and not section_exclusions:
            return original_fieldsets
        fieldsets = deepcopy(original_fieldsets)
        if section_exclusions:
            fieldsets = [(title, attributes) for title, attributes in fieldsets if title not in section_exclusions]
        if field_exclusions:
            remove_from_fieldsets(fieldsets, field_exclusions)
        return fieldsets

    def get_urls(self):
        urls = super().get_urls()
        my_urls = []

        def get_action_link_view(_view_name, _attrs):
            def action_link_view(request, pk):
                obj = self.model.objects.get(id=pk)
                response = _attrs['callback'](obj)
                if _attrs.get('custom_redirect'):
                    return response
                return redirect(request.META.get('HTTP_REFERER'))

            action_link_view.__name__ = _view_name
            return action_link_view

        for (name, attrs) in self.action_links:
            if 'callback' not in attrs:
                continue
            view_name = f'{self.model._meta.app_label}_{self.model._meta.model_name}_action_view_{name}'

            my_urls.append(path(f'<path:pk>/action-link/{name}/',
                                self.admin_site.admin_view(get_action_link_view(view_name, attrs)),
                                name=view_name))
        return my_urls + urls


class ReadonlyAdmin(admin.ModelAdmin):

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def get_readonly_fields(self, request, obj=None):
        return self.get_list_display(request)

    def __init__(self, model, admin_site):
        self.list_display = [field.name for field in model._meta.fields if field.name
                             not in {'id', 'created', 'modified'}]
        super().__init__(model, admin_site)
